#ifndef COUNTER_CONTAINER_INCLUDED__
#define COUNTER_CONTAINER_INCLUDED__

#include <map>

template <class T>
T *AddRef(T *pT)
{
   if (pT)
   {
      pT->AddRef();
   }

   return pT;
}

template <class T>
T *Release(T *pT)
{
   if (pT)
   {
      pT->Release();
   }

   return 0;
}


template <class T>
class TCounterContainer
{
   public :

      template <class T>
      class Node
      {
         public :

            Node(T *pData)
               :  m_refCount(1), 
                  m_pNext(0), 
                  m_pPrev(0),
                  m_pData(pData),
                  m_bActive(true)
            {
            }

            ~Node()
            {
               m_pNext = ::Release(m_pNext);
               m_pPrev = ::Release(m_pPrev);
               m_pData = ::Release(m_pData);
            }

            void SetNext(Node *pNext)
            {
               if (pNext)
               {
                  pNext->m_pPrev = ::AddRef(this);
               }

               ::Release(m_pNext);
               m_pNext = ::AddRef(pNext);
            }

            void AddRef()
            {
               ::InterlockedIncrement(&m_refCount);
            }

            void Release()
            {
               if (0 == ::InterlockedDecrement(&m_refCount))
               {
                  delete this;
               }
            }

            void Unlink()
            {
               m_bActive = false;

               if (m_pPrev)
               {
                  m_pPrev->SetNext(m_pNext);
               }
               else
               {
                  m_pNext->m_pPrev = ::Release(m_pNext->m_pPrev);
               }
            }

            bool Active()
            {
               return m_bActive;
            }

            Node *GetNext()
            {
               return m_pNext;
            }

            Node *GetPrev()
            {
               return m_pPrev;
            }

            T *Data()
            {
               return m_pData;
            }

         private :

            LONG m_refCount;

            Node *m_pNext;
            Node *m_pPrev;
            
            T *m_pData;

            bool m_bActive;
      };

      typedef std::map<std::string, Node<T> *> CounterMap;

      typedef CounterMap::const_iterator const_iterator;
      typedef CounterMap::iterator iterator;
      typedef std::pair<CounterMap::value_type::first_type, T *> value_type;
      typedef CounterMap::size_type size_type;

      const_iterator begin() const
      {
         return m_map.begin();
      }

      const_iterator end() const
      {
         return m_map.end();
      }

      iterator begin() 
      {
         return m_map.begin();
      }

      iterator end() 
      {
         return m_map.end();
      }

      iterator find(const std::string &name) 
      {
         return m_map.find(name);
      }
      
      std::pair<iterator, bool> insert(const value_type &value)
      {
         Node<T> *pNode = new Node<T>(value.second);

         std::pair<iterator, bool> result = m_map.insert(CounterMap::value_type(value.first, pNode));

         if (result.second == true)
         {
            AddNode(pNode);
         }

         return result;
      }

      iterator erase(iterator it)
      {
         Node<T> *pNode = (*it).second;
         
         if (m_pHead == pNode)
         {
            m_pHead = pNode->GetNext();
            ::AddRef(m_pHead);
         }
         
         if (m_pTail == pNode)
         {
            m_pTail = pNode->GetPrev();
            ::AddRef(m_pTail);
         }

         pNode->Unlink();
         pNode->Release();

         return m_map.erase(it);
      }

      size_type size() const
      {
         return m_map.size();
      }

      TCounterContainer()
         : m_pHead(0), m_pTail(0)
      {
      }

      ~TCounterContainer()
      {
         Release(m_pHead);
         Release(m_pTail);
      }

      template <class T>
      class Tlist_iterator
      {
         public :

            Tlist_iterator(Node<T> *pNode)
               : m_pNode(pNode)
            {
               if (m_pNode)
               {
                  m_pNode->AddRef();
               }
            }

            Tlist_iterator(const Tlist_iterator<T> &rhs)
               : m_pNode(::AddRef(rhs.m_pNode))
            {
            }

            ~Tlist_iterator()
            {
               m_pNode = ::Release(m_pNode);
            }

            Tlist_iterator &operator =(const Tlist_iterator<T> &rhs)
            {
               if (this != &rhs)
               {
                  ::Release(m_pNode);
                  m_pNode = ::AddRef(rhs.m_pNode);
               }

               return *this;
            }

            bool operator==(const Tlist_iterator<T> &rhs)
            {
               CheckNodeActive();
               
               rhs.CheckNodeActive();

               return (m_pNode == 0 && rhs.m_pNode == 0);
            }

            bool operator!=(const Tlist_iterator<T> &rhs)
            {
               return !(*this == rhs);
            }


            T *operator*() const
            {
               CheckNodeActive();

               if (!m_pNode)
               {
                  throw "TODO";
               }

               return m_pNode->Data();
            }


            T &operator->() const
            {
               CheckNodeActive();

               if (!m_pNode)
               {
                  throw "TODO";
               }

               return *(m_pNode->Data());
            }

            Tlist_iterator<T> &operator++()
            {
               
               if (m_pNode)
               {
                  CheckNodeActive();

                  NextNode();
               }
               
               return *this;
            }

         private :

            void CheckNodeActive() const
            {
               while (m_pNode && !m_pNode->Active())
               {
                  NextNode();
               }
            }

            void NextNode() const
            {
               Node<T> *pNext = m_pNode->GetNext();
               
               ::AddRef(pNext);

               ::Release(m_pNode);
               
               m_pNode = pNext;
            }

            mutable Node<T> *m_pNode;
      };

      typedef Tlist_iterator<T> list_iterator;


      list_iterator beginList() 
      {
         return list_iterator(m_pHead);
      }

      list_iterator endList() 
      {
         return list_iterator(0);
      }

   private :


      void AddNode(Node<T> *pNode)
      {
         if (!m_pHead)
         {
            m_pHead = pNode;
            m_pTail = ::AddRef(pNode);
         }
         else 
         {
            m_pTail->SetNext(pNode);
            m_pTail = pNode;
         }
      }

      Node<T> *m_pHead;
      Node<T> *m_pTail;


      CounterMap m_map;
};

#endif // COUNTER_CONTAINER_INCLUDED__