#include "LRUList.h"

#include <string>
#include <strstream>
#include <time.h>

#define DEBUG_TRACE

#ifdef DEBUG_TRACE
#include <iostream>
#define DEBUG_ONLY(a) a
#else
#define DEBUG_ONLY(a) 
#endif

class LRUList::UsageMonitor
{
   public :

      UsageMonitor()
         :  m_pServant(0), 
            m_creationTime(0), 
            m_lastUseTime(0) 
      {
      }

      void SetServant(
         PortableServer::RefCountServantBase *pServant)
      {
         m_pServant = pServant;
         m_creationTime = time(0);     // maybe use GetTickCount() on nt..
      }

      PortableServer::RefCountServantBase *GetServant() const
      {
         return m_pServant;
      }

      PortableServer::RefCountServantBase *GetServant() 
      {
         m_lastUseTime = time(0);
         
         return m_pServant;
      }

      void Clear()
      {
         m_pServant = 0;
         m_creationTime = 0;
         m_lastUseTime = 0;
      }

      time_t GetCreationTime() const
      {
         return m_creationTime;
      }

      time_t GetUsageTime() const
      {
         return m_lastUseTime;
      }

   private :

      PortableServer::RefCountServantBase *m_pServant;
      time_t m_creationTime;
      time_t m_lastUseTime;
};


LRUList::LRUList(
   const size_t maxServants,
   const size_t timeout)
   :  m_numServants(0),
      m_maxServants(maxServants),
      m_timeout(timeout)
{
}

LRUList::~LRUList()
{
}


bool LRUList::TimedOut(const size_t index) const
{
   return (time(0) - m_servants[index].GetUsageTime() > m_timeout);
}
      
void LRUList::Evict()
{
   bool done = m_lruList.empty();
   
   while(!done)
   {
      const size_t index = m_lruList.front();

      if (TimedOut(index))
      {
         std::cout << "Evict due to timeout" << std::endl;
         
         m_servants[index].GetServant()->_remove_ref();

         m_lruList.pop_front();

         done = m_lruList.empty();
      }
      else
      {
         done = true;
      }
   }
}

PortableServer::ObjectId_var LRUList::Add(PortableServer::RefCountServantBase *pNewObject)
{
   int index;

   Evict();       // force eviction if servants have timed out.

   if (!m_freeList.empty())
   {
      index = m_freeList.front();
      m_freeList.pop_front();
   }
   else if (m_numServants < m_maxServants)
   {
      std::cout << "Create new node" << std::endl;

      UsageMonitor mon;

      m_servants.push_back(mon);

      index = m_servants.size() - 1;

      m_numServants++;
   }
   else
   {
      // what to do when we run out of space?

      throw CORBA::IMP_LIMIT();
   }

   m_servants[index].SetServant(pNewObject);

   PortableServer::ObjectId_var oid = GenerateOid(index);

   UseServant(index);

   DisplayList();

   return oid;
}

void LRUList::Remove(const PortableServer::ObjectId &oid)
{
   const size_t index = OidToIndex(oid);

   m_servants[index].Clear();

      need to have seperate ref count because we must be able to destroy an object
      with outstanding refs

   m_lruList.remove(index);

   m_freeList.push_back(index);

//   DisplayList();
}

PortableServer::RefCountServantBase *LRUList::operator[] (const PortableServer::ObjectId &oid) 
{
   const size_t index = OidToIndex(oid);

   return UseServant(index);
}

size_t LRUList::GetTimeout() const
{
   return m_timeout;
}

PortableServer::RefCountServantBase *LRUList::UseServant(const size_t index)
{
   if (m_lruList.empty() || m_lruList.back() != index)
   {
      m_lruList.remove(index);
      m_lruList.push_back(index);
   }
     
   return m_servants[index].GetServant();
}

void LRUList::DisplayList() const
{
   std::cout << "----- List is: -----" << std::endl;
   std::cout << "Num Nodes = " << m_numServants << std::endl;
   std::cout << "Max Nodes = " << m_maxServants << std::endl;

   std::list<size_t>::const_iterator it;

   for (it = m_lruList.begin(); it != m_lruList.end(); ++it)
   {
      const size_t index = *it;

      std::cout << index << " - " <<  "[" << m_servants[index].GetServant();
      std::cout << "] oid = [" << index << " " << m_servants[index].GetCreationTime();
      std::cout << "] " << m_servants[index].GetUsageTime() << std::endl;
   }
   
   std::cout << "---- Free List: ----" << std::endl;

   for (it = m_freeList.begin(); it != m_freeList.end(); ++it)
   {
      std::cout << *it << std::endl;
   }

   std::cout << "--------------------" << std::endl;

}

PortableServer::ObjectId_var LRUList::GenerateOid(
   const size_t index) const
{
   // Generate an OID based on the index and the time the servant was created

   std::strstream stream;

   stream << (index + 1) << " " << m_servants[index].GetCreationTime() << std::ends;

   std::string result = stream.str();

   stream.rdbuf()->freeze(false);

   return PortableServer::string_to_ObjectId(result.c_str());
}

size_t LRUList::OidToIndex(const PortableServer::ObjectId &oid) const
{
	CORBA::String_var oid_string = PortableServer::ObjectId_to_string(oid);

   char *pOidStr = oid_string;

   size_t index = atol(strtok(pOidStr, " "));

   time_t oid_time = atol(strtok(0, " "));

   if (index > m_servants.size() || index == 0)
   {
      throw CORBA::OBJECT_NOT_EXIST();
   }

   index = index - 1;

   if (m_servants[index].GetCreationTime() != oid_time)
   {
      throw CORBA::OBJECT_NOT_EXIST();
   }

   return index;
}
